import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

public class LensCorrectionFilter extends EffectFilter
{	
	// lens distortion parameters
	private double ap;
	private double bp;
	private double cp;
	private double dp;
	
	// transform table size
	private int imgWidth;
	private int imgHeight;
	
	// calculated transformation table
	private IntPoint[][] transformTable;
	
	// fields for entering the parameters
	protected transient JFormattedTextField aField;
	protected transient JFormattedTextField bField;
	protected transient JFormattedTextField cField;
	protected transient JFormattedTextField dField;

	public LensCorrectionFilter()
	{
		filterInputs = new FilterLink[1];
		filterInputs[0] = null;
		dead = false;
		filterName = "Lens Correction Filter " + (int)(Math.random() * 1000);
		
		guiWidth = 100;
		guiHeight = 100;
		guiX = 150;
		guiY = 150;
		
		ap = 0.001;
		bp = -0.001;
		cp = 0.001;
		dp = -0.001;
		
		rebuildFromSave();
	}
	
	public void prepareForSave()
	{
		super.prepareForSave();
	}
	
	public void rebuildFromSave()
	{
		internalBuffer = new FloatImage();
		needsProcessing = true;
		
		guiWindow = new JInternalFrame(filterName, true, true, true, true);
		guiWindow.setSize(guiWidth, guiHeight);
		guiWindow.setLocation(guiX, guiY);
		
		JPanel foo = new JPanel();
		
		aField = new JFormattedTextField(new Double(ap));
		aField.setBorder(BorderFactory.createTitledBorder("a"));
		aField.setPreferredSize(new Dimension(200, 50));
		
		bField = new JFormattedTextField(new Double(bp));
		bField.setBorder(BorderFactory.createTitledBorder("b"));
		bField.setPreferredSize(new Dimension(200, 50));
		
		cField = new JFormattedTextField(new Double(cp));
		cField.setBorder(BorderFactory.createTitledBorder("c"));
		cField.setPreferredSize(new Dimension(200, 50));
		
		dField = new JFormattedTextField(new Double(dp));
		dField.setBorder(BorderFactory.createTitledBorder("d"));
		dField.setPreferredSize(new Dimension(200, 50));
		
		foo.add(aField);
		foo.add(bField);
		foo.add(cField);
		foo.add(dField);
		
		JButton calculateButton = new JButton("Make Transform");
		calculateButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e)
			{
				doTableCalculation();
			}
		});
		foo.add(calculateButton);
		
		guiWindow.add(foo);
		
		guiWindow.setVisible(true);
	}
	
	public String getName()
	{
		return filterName;
	}
	
	public String getInputName(int inputNum)
	{
		return "Input";
	}
	
	public String getOutputName(int outputNum)
	{
		return "Output";
	}
	
	public int numOutputs()
	{
		return 1;
	}
	
	public void doTableCalculation()
	{
		transformTable = new IntPoint[imgWidth][imgHeight];
		
		ap = (Double)aField.getValue();
		bp = (Double)bField.getValue();
		cp = (Double)cField.getValue();
		dp = (Double)dField.getValue();
		
		calculateTable();
	}
	
	public void calculateTable()
	{
		double cx = (double)imgWidth / 2.0;
		double cy = (double)imgHeight / 2.0;
		
		// normalized radius: radiusn = 1.0
		double radiusn = imgHeight / 2.0;
		
		double sr, dr, dr2, dr3, dr4, scaleFactor, dx, dy;
		int ti, tj;
		
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{
				dx = (double)i - cx;
				dy = (double)j - cy;
				
				dr = Math.sqrt((dx * dx) + (dy * dy)) / radiusn;
				dr2 = dr * dr;
				dr3 = dr * dr2;
				dr4 = dr2 * dr2;
				
				sr = (ap * dr4) + (bp * dr3) + (cp * dr2) + (dp * dr);
				
				if(dr > 0)
					scaleFactor = sr / dr;
				else
					scaleFactor = 0;
				
				ti = (int)(cx + (dx * scaleFactor));
				tj = (int)(cy + (dy * scaleFactor));
			
				transformTable[i][j] = new IntPoint(0,0);
				
				if(ti >= 0 && ti < imgWidth && tj >= 0 && tj < imgHeight)
				{
					transformTable[i][j].x = ti;
					transformTable[i][j].y = tj;
				}
				
			}
		}
	}
	
	private FloatImage doTransform()
	{
		if(needsProcessing && filterInputs[0] != null)
		{
			FloatImage temp = filterInputs[0].getOutput();
			
			if(temp != null)
			{	
				if(temp.getWidth() != internalBuffer.getWidth() || temp.getHeight() != internalBuffer.getHeight())
				{
					internalBuffer = new FloatImage(3, temp.getWidth(), temp.getHeight());
					
					// if the input size changes, it invalidates the transform table
					transformTable = null;
				}
				
				imgWidth = internalBuffer.getWidth();
				imgHeight = internalBuffer.getHeight();
				
				double[][] ored = internalBuffer.getChannelData(0);
				double[][] ogreen = internalBuffer.getChannelData(1);
				double[][] oblue = internalBuffer.getChannelData(2);
				
				double[][] ired = temp.getChannelData(0);
				double[][] igreen = temp.getChannelData(1);
				double[][] iblue = temp.getChannelData(2);
				
				if(transformTable != null)
				{
					int ti, tj;
					
					for(int i = 0; i < imgWidth; ++i)
					{
						for(int j = 0; j < imgHeight; ++j)
						{
							if(transformTable[i][j] != null)
							{
								ti = transformTable[i][j].x;
								tj = transformTable[i][j].y;
								
								if(ti >= 0 && tj >= 0 && ti < imgWidth && tj < imgHeight)
								{
									ored[i][j] = ired[ti][tj];
									ogreen[i][j] = igreen[ti][tj];
									oblue[i][j] = iblue[ti][tj];
								}
							}
						}
					}
				}
				else // transformTable == null
				{
					for(int i = 0; i < imgWidth; ++i)
					{
						for(int j = 0; j < imgHeight; ++j)
						{
							ored[i][j] = ired[i][j];
							ogreen[i][j] = igreen[i][j];
							oblue[i][j] = iblue[i][j];
						}
					}
				}
			}
			needsProcessing = false;
		}

		return internalBuffer;
	}
	
	public FloatImage getOutput(int outputChannel)
	{	
		return doTransform();
	}	
}