import java.util.*;

public class OUtil
{
	public static class Edge
	{
		public double x1;
		public double x2;
		public double y1;
		public double y2;
		
		public boolean used;
		
		public Edge()
		{
			used = false;
		}
		
		public Edge(double nx1, double ny1, double nx2, double ny2)
		{
			x1 = nx1;
			y1 = ny1;
			x2 = nx2;
			y2 = ny2;
			
			/*  p1 (x1, y1) is always above or to the left of p2  
			 * 
			 *  p1      p2
			 *    _______
			 *   |
			 *   |
			 *   |
			 *   |p2
			 */
			
			used = false;
		}
		
		public Point getOtherPoint(double x, double y)
		{
			if(Math.abs(x1 - x) + Math.abs(y1 - y) < Math.abs(x2 - x) + Math.abs(y2 - y))
			{
				return new Point(x2, y2);
			}
			else
			{
				return new Point(x1, y1);
			}
		}
		
		public Point getOtherPoint(Point first)
		{
			if(first != null)
				return getOtherPoint(first.x, first.y);
			else
				return null;
		}
	}
	
	public static class Pixel
	{
		Edge[] edges;
		
		/*    __e0_
		 *   |     |
		 * e3|     |e1
		 *   |_____|
		 *      e2
		 */
		
		public Pixel()
		{
			edges = new Edge[4];
		}
		
		public int numEdges()
		{
			int num = 0;
			
			for(int i = 0; i < 4; ++i)
			{
				if(edges[i] != null)
					++num;
			}
			
			return num;
		}
	}
	
	public static class Point
	{
		public double x;
		public double y;
		
		public int arcPos;
		
		public Point(double nx, double ny)
		{
			x = nx; 
			y = ny;
			
			arcPos = -1;
		}
		
		public Point(double nx, double ny, int nArcPos)
		{
			x = nx;
			y = ny;
			arcPos = nArcPos;
		}
		
		public Point(Point src)
		{
			x = src.x;
			y = src.y;
			arcPos = src.arcPos;
		}
		
	}
	
	public static class Cut
	{
		Point p1;
		Point p2;
		double length;
		
		public Cut(Point a, Point b)
		{
			p1 = a;
			p2 = b;
			
			double dx = p2.x - p1.x;
			double dy = p2.y - p1.y;
			
			length = Math.sqrt(dx * dx + dy * dy);
		}
		
		// note: equals only makes sure that the positions along the contour
		//  of the two cuts are the same, not that their x and y coordinates
		//  are also identical
		public boolean equals(Object rhs)
		{
			if(rhs == null)
				return false;
			
			if(rhs instanceof Cut)
			{
				Cut crhs = (Cut)rhs;
				return (p1.arcPos == crhs.p1.arcPos && p2.arcPos == crhs.p2.arcPos);
			}
			else
			{
				return false;
			}
		}
	}
	
	public static class Segment
	{
		int startPos;
		int endPos;
		int owner;
		
		public Segment()
		{
			startPos = -1;
			endPos = -1;
			owner = -1;
		}
		
		public Segment(int s, int e, int o)
		{
			startPos = s;
			endPos = e;
			owner = o;
		}
		
		public Segment(Segment src)
		{
			startPos = src.startPos;
			endPos = src.endPos;
			owner = src.owner;
		}
		
		public boolean containsPos(int pos)
		{
			if(startPos <= endPos)
			{
				return (pos >= startPos && pos <= endPos);
			}
			else // startPos > endPos
			{
				return (pos >= startPos || pos <= endPos);
			}
		}
	}
	
	public static class BoundingBox implements Comparable
	{
		double top;
		double left;
		double bottom;
		double right;
		
		double width;
		double height;
		
		boolean marked;
		
		Contour contour;
		BoundingBox match;
		
		public BoundingBox(Contour contour)
		{
			top = 10000.0;
			left = 10000.0;
			right = -1.0;
			bottom = -1.0;
			
			match = null;
			marked = false;
			
			Point p;
			
			for(int i = 0; i < contour.path.size(); ++i)
			{
				p = contour.path.get(i);
				
				if(p.x < left)
					left = p.x;
				
				if(p.x > right)
					right = p.x;
					
				if(p.y < top)
					top = p.y;
					
				if(p.y > bottom)
					bottom = p.y;
			}
			
			width = right - left;
			height = bottom - top;
		}
		
		public BoundingBox()
		{
			match = null;
			marked = false;
		}
		
		public void mark()
		{
			marked = true;
		}
		
		public void unmark()
		{
			marked = false;
		}
		
		public int compareTo(Object o)
		{
			BoundingBox rhs = (BoundingBox)o;
			
			if(rhs.top > top)
				return -1;
			else if(rhs.top < top)
				return 1;
			else
				return 0;
		}
		
		
	}
	
	public static class BodyComponent
	{
		int type;
		ArrayList<BodyComponent> links;
		ArrayList<Cut> otherCuts;
		Cut primaryCut;
		Contour contour;
		
		// center of mass location
		double cmx;
		double cmy;
		
		double top;
		double left;
		double bottom;
		double right;
		
		public BodyComponent()
		{
			contour = new Contour();
			primaryCut = null;
			links = new ArrayList<BodyComponent>();
			otherCuts = new ArrayList<Cut>();
			type = 0;
			
			cmx = -1.0;
			cmy = -1.0;
			
			top = -1.0;
			left = -1.0;
			bottom = -1.0;
			right = -1.0;
		}
		
		public double gcmx()
		{
			if(cmx < 0)
				calculateCM();
				
			return cmx;
		}
		
		public double gcmy()
		{
			if(cmy < 0)
				calculateCM();
				
			return cmy;
		}
		
		public int degree()
		{
			return links.size();
		}
		
		// these return the bounding box for the component
		public double bbtop()
		{
			if(top < 0)
				calculateCM();
				
			return top;
		}
		
		public double bbleft()
		{
			if(left < 0)
				calculateCM();
				
			return left;
		}
		
		public double bbright()
		{
			if(right < 0)
				calculateCM();
				
			return right;
		}
		
		public double bbbottom()
		{
			if(bottom < 0)
				calculateCM();
				
			return bottom;
		}
		
		public void calculateCM()
		{
			double sumx = 0.0;
			double sumy = 0.0;
			
			top = 10000.0;
			left = 10000.0;
			right = -1.0;
			bottom = -1.0;
			
			Point p;
			
			for(int i = 0; i < contour.path.size(); ++i)
			{
				p = contour.path.get(i);
				sumx += p.x;
				sumy += p.y;
				
				if(p.x < left)
					left = p.x;
				
				if(p.x > right)
					right = p.x;
					
				if(p.y < top)
					top = p.y;
					
				if(p.y > bottom)
					bottom = p.y;
			}
			
			cmx = sumx / contour.path.size();
			cmy = sumy / contour.path.size();
		}
		
		public void drawLinks(FloatImage dest)
		{
			BodyComponent bc;
			
			double[][] ogreen = dest.getChannelData(1);
			
			ogreen[(int)gcmx()][(int)gcmy()] = 1.0; 
			
			
			for(int i = 0; i < links.size(); ++i)
			{
				bc = links.get(i);
				
				drawLine(dest, (int)gcmx(), (int)(bc.gcmx()), (int)gcmy(), (int)(bc.gcmy()), 1.0);
			}
		}
	}
	
	public static class Contour
	{
		ArrayList<Point> path;
		
		public Contour()
		{
			path = new ArrayList<Point>();
		}
		
		public Contour(ArrayList<Point> npath)
		{
			path = npath;
		}
	}
	
	public static class EdgeField
	{
		Pixel[][] efield;
		ArrayList<Edge> allEdges;
		int fieldWidth;
		int fieldHeight;
		
		public EdgeField()
		{
			efield = null;
			allEdges = null;
			fieldWidth = 0;
			fieldHeight = 0;
		}
		
		public ArrayList<Contour> getOutlines()
		{
			ArrayList<Point> currentPath;
			ArrayList<Contour> retVal = new ArrayList<Contour>();
			
			Edge curEdge;
			Point curPoint;
			Pixel curPixel1, curPixel2;
			int ix, iy, tally;
			
			if(allEdges.size() == 0)
				System.out.println("No edges!");
				
			//System.out.println("Num edges: " + allEdges.size());
			
			for(int i = 0; i < allEdges.size(); ++i)
			{
				curEdge = allEdges.get(i);
				
				if(!curEdge.used)
				{
					// follow this edge around
					currentPath = new ArrayList<Point>();
					
					curPoint = new Point(curEdge.x1, curEdge.y1);
					
					while(!curEdge.used)
					{
						currentPath.add(curPoint);
						curEdge.used = true;
						
						curPoint = curEdge.getOtherPoint(curPoint);
						
						//System.out.println("Looking at " + curPoint.x + ", " + curPoint.y);
						
						//System.out.println("Int: " + (int)curPoint.x + ", " + (int)curPoint.y);
						
						ix = (int)curPoint.x;
						iy = (int)curPoint.y;
						
						if(ix > 0 && iy > 0)
						{
							curPixel1 = efield[ix][iy];
							curPixel2 = efield[ix - 1][iy - 1];
							
							/*
							tally = 0;
							
							if(curPixel1.edges[0] != null)
								++tally;
							if(curPixel1.edges[3] != null)
								++tally;
							if(curPixel2.edges[1] != null)
								++tally;
							if(curPixel2.edges[2] != null)
								++tally;
							
							System.out.println("Num edges: " + tally);
							*/
							
							if(curPixel1.edges[0] != null && !curPixel1.edges[0].used)
							{
								//System.out.println("Gotcha! " + curPoint.x + ", " + curPoint.y);
								curEdge = curPixel1.edges[0];
							}
							else if(curPixel2.edges[1] != null && !curPixel2.edges[1].used)
							{
								//System.out.println("Gotcha! " + curPoint.x + ", " + curPoint.y);
								curEdge = curPixel2.edges[1];
							}
							else if(curPixel2.edges[2] != null && !curPixel2.edges[2].used)
							{
								//System.out.println("Gotcha! " + curPoint.x + ", " + curPoint.y);
								curEdge = curPixel2.edges[2];
							}
							else if(curPixel1.edges[3] != null && !curPixel1.edges[3].used)
							{
								//System.out.println("Gotcha! " + curPoint.x + ", " + curPoint.y);
								curEdge = curPixel1.edges[3];
							}
						}
					}
					
					//System.out.println("----------");
					
					retVal.add(new Contour(currentPath));
				}
			}
			return retVal;
		}
		
		public void makeEdgeField(FloatImage src)
		{
			if(src == null)
				return;
			
			int srcWidth = src.getWidth();
			int srcHeight = src.getHeight();
				
			if(efield == null || fieldWidth != srcWidth || fieldHeight != srcHeight)
			{
				efield = new Pixel[srcWidth][srcHeight];
				fieldWidth = srcWidth;
				fieldHeight = srcHeight;
				
				// first, make all the pixels
				for(int i = 0; i < fieldWidth; ++i)
				{
					for(int j = 0; j < fieldHeight; ++j)
					{
						efield[i][j] = new Pixel();
					}
				}	
			}
			
			// assume that the image is a binary image identical in all channels, so
			//  that we only have to deal with one channel
			double[][] srcdata = src.getChannelData(0);
			
			allEdges = new ArrayList<Edge>();
			
			// we will assume that the object's contour does not touch the edges of
			// the image (make a 1 pixel black border all around the image before
			// using makeEdgeField)
			
			Edge temp;
			
			//System.out.println("Made empty efield!");
			
			// now make the edges
			for(int i = 1; i < fieldWidth - 1; ++i)
			{
				for(int j = 1; j < fieldHeight - 1; ++j)
				{	
					// if this pixel is not white, don't bother to do anything at
					//  all
					
					if(srcdata[i][j] > 0.9)
					{
						//System.out.println("Found white");
						
						// now we consider each of the four possible edges in turn
						//  (we add the edge if the neighbor is black: since we only
						//  consider white pixels, this ensures that the edge is only
						//  added once)
						
						// Left (e3 for this pixel, e1 for left pixel)
						if(srcdata[i-1][j] < 0.1)
						{
							temp = new Edge(i, j, i, j+1);
							
							efield[i][j].edges[3] = temp;
							efield[i-1][j].edges[1] = temp;
							
							allEdges.add(temp);
							
							//System.out.println("Added edge!");
						}
						
						// Right (e1 for this, e3 for right)
						if(srcdata[i+1][j] < 0.1)
						{
							temp = new Edge(i+1, j, i+1, j+1);
							
							efield[i][j].edges[1] = temp;
							efield[i+1][j].edges[3] = temp;
							
							allEdges.add(temp);
							
							//System.out.println("Added edge!");
						}
						
						// Top (e0 for this, e2 for top)
						if(srcdata[i][j-1] < 0.1)
						{
							temp = new Edge(i, j, i+1, j);
							
							efield[i][j].edges[0] = temp;
							efield[i][j-1].edges[2] = temp;
							
							allEdges.add(temp);
							
							//System.out.println("Added edge!");
						}
						
						// Bottom (e2 for this, e0 for bottom)
						if(srcdata[i][j+1] < 0.1)
						{
							temp = new Edge(i, j+1, i+1, j+1);
							
							efield[i][j].edges[2] = temp;
							efield[i][j+1].edges[0] = temp;
							
							allEdges.add(temp);
							
							//System.out.println("Added edge!");
						}
					}
				}
			}
		}
	}
	
	public static Point calculateCM(Contour contour)
	{
		double sumx = 0.0;
		double sumy = 0.0;
		
		Point p;
		
		for(int i = 0; i < contour.path.size(); ++i)
		{
			p = contour.path.get(i);
			sumx += p.x;
			sumy += p.y;
		}
		
		return new Point(sumx / contour.path.size(), sumy / contour.path.size());
	}
	
	public static Contour growContour(Contour src, double factor)
	{
		Point cm = calculateCM(src);
		
		Point p;

		Contour ret = new Contour();
		
		double dx, dy;

		for(int i = 0; i < src.path.size(); ++i)
		{
			p = src.path.get(i);
			
			dx = p.x - cm.x;
			dy = p.y - cm.y;
			
			ret.path.add(new Point(cm.x + dx * factor, cm.y + dy * factor, p.arcPos));
		}
		
		return ret;
	}
	
	public static Double[] calculateCurvature(Contour src)
	{
		if(src == null)
			return null;
			
		int numPoints = src.path.size();
		
		// we need at least 3 points
		if(numPoints < 3)
			return null;
		
		Double[] xp, yp, xpp, ypp, kurv;
		
		xp = new Double[numPoints];
		yp = new Double[numPoints];
		xpp = new Double[numPoints];
		ypp = new Double[numPoints];
		kurv = new Double[numPoints];
		
		Point curPoint1 = null;
		Point curPoint2 = null;
		
		int curIndex1 = 0;
		int curIndex2 = 0;
		
		// first we calculate the first derivatives
		//  (we have to calculate the first and last points as special cases
		//   because the path loops around)
		
		// first point
		curPoint1 = src.path.get(numPoints-1);
		curPoint2 = src.path.get(1);
		
		xp[0] = 0.5 * (curPoint2.x - curPoint1.x);
		yp[0] = 0.5 * (curPoint2.y - curPoint1.y);
		
		// now all the points except the last
		for(int i = 1; i < numPoints-1; ++i)
		{
			curPoint1 = src.path.get(i-1);
			curPoint2 = src.path.get(i+1);
			
			xp[i] = 0.5 * (curPoint2.x - curPoint1.x);
			yp[i] = 0.5 * (curPoint2.y - curPoint1.y);
		}
		
		// now the last point
		curPoint1 = src.path.get(numPoints-2);
		curPoint2 = src.path.get(0);
		
		xp[numPoints-1] = 0.5 * (curPoint2.x - curPoint1.x);
		yp[numPoints-1] = 0.5 * (curPoint2.y - curPoint1.y);
		
		////////////////////////////////////////////////
		// now second derivates in the identical fashion
		/////////////////////////////////////////////////
		
		curIndex1 = numPoints-1;
		curIndex2 = 1;
		
		xpp[0] = 0.5 * (xp[curIndex2] - xp[curIndex1]);
		ypp[0] = 0.5 * (yp[curIndex2] - yp[curIndex1]);
		
		// now all the points except the last
		for(int i = 1; i < numPoints-1; ++i)
		{
			curIndex1 = i-1;
			curIndex2 = i+1;
			
			xpp[i] = 0.5 * (xp[curIndex2] - xp[curIndex1]);
			ypp[i] = 0.5 * (yp[curIndex2] - yp[curIndex1]);
		}
		
		// now the last point
		curIndex1 = numPoints-2;
		curIndex2 = 0;
		
		xpp[numPoints-1] = 0.5 * (xp[curIndex2] - xp[curIndex1]);
		ypp[numPoints-1] = 0.5 * (yp[curIndex2] - yp[curIndex1]);
		
		// finally we compute the curvature
		// k = xp * ypp - yp * xpp
		//     -------------------
		//     (xp^2 + yp^2)^(3/2)
		
		//double k = 0.0;
		double num = 0.0;
		double sqpart = 0.0;
		double den = 0.0;
		
		for(int i = 0; i < numPoints; ++i)
		{
			num = (xp[i] * ypp[i]) - (yp[i] * xpp[i]);
			sqpart = (xp[i] * xp[i]) + (yp[i] * yp[i]);
			den = Math.sqrt(sqpart * sqpart * sqpart);
			
			if(den > 0.0)
				kurv[i] = num / den;
			else
				kurv[i] = 1000.0;	
		}
		
		return kurv;
	}
	
	public static ArrayList<Point> identifyCriticalPoints(Contour src, Double[] kurv, int rad, double cThreshold)
	{
		final double THRESHOLD = 0.02;
		
		if(kurv == null)
			kurv = calculateCurvature(src);
			
		if(kurv == null || src == null)
			return null;
		
		int numPoints = src.path.size();
		
		int srcLoc;
		
		Point temp;
		
		boolean isCritical = false;
		
		ArrayList<Point> crits = new ArrayList<Point>();
		
		for(int i = 0; i < numPoints; ++i)
		{
			isCritical = (kurv[i] > cThreshold);
			
			for(int k = i - rad; k <= i + rad; ++k)
			{
				srcLoc = k;
				
				while(srcLoc < 0)
				{
					srcLoc += numPoints;
				}
					
				if(srcLoc >= numPoints)
					srcLoc = srcLoc % numPoints;
					
				if(kurv[srcLoc] > kurv[i])
					isCritical = false;	
			}
			
			if(isCritical)
			{
				temp = src.path.get(i);
				temp.arcPos = i;
				crits.add(temp);
			}
		}
		
		return crits;
	}
	
	public static Contour smoothContour(Contour src, int smoothAmount)
	{
		if(src == null || src.path.size() < smoothAmount)
			return null;
		
		ArrayList<Point> smoothed = new ArrayList<Point>(src.path.size());
		
		double numAvg = (smoothAmount * 2);
		
		double sumX, sumY;
		
		int srcSize = src.path.size();
		
		int srcLoc;
		
		for(int i = 0; i < srcSize; ++i)
		{
			sumX = 0.0;
			sumY = 0.0;
			
			for(int k = i-smoothAmount; k < i+smoothAmount; ++k)
			{
				srcLoc = k;
				
				// this is to deal with the circular path: negative indices in
				//  the path should just loop back around to the positive
				//  (we can't just use %, because (for instance) -10 % 6 = -4
				//  rather than -10 % 6 = 2
				
				while(srcLoc < 0)
				{
					srcLoc += srcSize;
				}
				
				if(srcLoc >= srcSize)
					srcLoc = srcLoc % srcSize;
				
				sumX += src.path.get(srcLoc).x;
				sumY += src.path.get(srcLoc).y;
			}
			
			smoothed.add(new Point(sumX/numAvg, sumY/numAvg));
		}
		
		return new Contour(smoothed);
	}
	
	public static Contour downsampleContour(Contour src, int samplePoints)
	{
		if(src == null)
			return null;
		
		ArrayList<Point> sampled = new ArrayList<Point>();
		
		double sumX, sumY;
		
		double sa = (double)samplePoints;
		
		int srcSize = src.path.size();
		
		int srcLoc;
		
		for(int i = 0; i < srcSize; ++i)
		{
			sumX = 0.0;
			sumY = 0.0;
			
			for(int k = i; k < i+samplePoints; ++k)
			{
				srcLoc = k;
				
				if(srcLoc >= srcSize)
					srcLoc = srcLoc % srcSize;
				
				sumX += src.path.get(srcLoc).x;
				sumY += src.path.get(srcLoc).y;
			}
			
			sampled.add(new Point(sumX / sa, sumY / sa));
		}
		
		return new Contour(sampled);
	}
	
	public static void drawLine(FloatImage dest, int x1, int x2, int y1, int y2, double val)
	{
		double dx = x2 - x1;
		double dy = y2 - y1;
		int numSteps = (int)Math.max(Math.abs(dx), Math.abs(dy));
		
		if(x1 < 0 || x2 < 0 || x1 >= dest.getWidth() || x2 >= dest.getWidth() 
			|| y1 < 0 || y2 < 0 || y1 >= dest.getHeight() || y2 >= dest.getHeight())
			return;
		
		double sx = dx / (double)numSteps;
		double sy = dy / (double)numSteps;
		
		double cx = x1;
		double cy = y1;
		
		int ix, iy;
		
		double[][] ored 	= dest.getChannelData(0);
		double[][] ogreen 	= dest.getChannelData(1);
		double[][] oblue 	= dest.getChannelData(2);
		
		for(int i = 0; i < numSteps; ++i)
		{
			ix = (int)cx;
			iy = (int)cy;
			
			ored[ix][iy] = val;
			ogreen[ix][iy] = val;
			oblue[ix][iy] = val;
			
			cx += sx;
			cy += sy;
		}
	}
	
	public static void drawFilledBox(FloatImage dest, int x1, int y1, int x2, int y2, double r, double g, double b)
	{
		if(x1 < 0)
			x1 = 0;
		if(x1 >= dest.getWidth())
			x1 = dest.getWidth() - 1;
		if(x2 < 0)
			x2 = 0;
		if(x2 >= dest.getWidth())
			x2 = dest.getWidth() - 1;
		if(y1 < 0)
			y1 = 0;
		if(y1 >= dest.getHeight())
			y1 = dest.getHeight() - 1;
		if(y2 < 0)
			y2 = 0;
		if(y2 >= dest.getHeight())
			y2 = dest.getHeight() - 1;
			
		double[][] ored = dest.getChannelData(0);
		double[][] ogreen = dest.getChannelData(1);
		double[][] oblue = dest.getChannelData(2);
			
		for(int j = y1; j <= y2; ++j)
		{
			for(int i = x1; i <= x2; ++i)
			{
				ored[i][j] = r;
				ogreen[i][j] = g;
				oblue[i][j] = b;
			}
		}
		
	}
	
	public static boolean segmentIntersection(Point pa1, Point pa2, Point pb1, Point pb2)
	{
		Point swap;
		
		// first, make sure that the points are arranged with both p1 points to
		//  the left of the p2 points
		
		if(pa1.x > pa2.x)
		{
			swap = pa1;
			pa1 = pa2;
			pa2 = swap;
		}
		
		if(pb1.x > pb2.x)
		{
			swap = pb1;
			pb1 = pb2;
			pb2 = swap;
		}
		
		// now we figure out the region of possible intersection
		double startx, endx;
		
		// the region of intersection starts (if anywhere) at the later x 
		//  coordinate of the two segments
		startx = Math.max(pa1.x, pb1.x);
		
		// the intersection region ends at the earlier of the x coordinates of the
		//  second points of the two segments
		endx = Math.min(pa2.x, pb2.x);
		
		// if the region is 'backwards', no intersection is possible
		if(startx > endx)
			return false;
			
		// now we must compute the y-coordinates of both segments at both the
		//  start and end
		double sya, syb, eya, eyb;
		double dxa, dya, dxb, dyb;
		
		dxa = pa2.x - pa1.x;
		dya = pa2.y - pa1.y;
		
		dxb = pb2.x - pb1.x;
		dyb = pb2.y - pb1.y;
		
		if(Math.abs(dxa) < 0.01)
		{
			sya = pa1.y;
			eya = pa2.y;
		}
		else
		{
			sya = (((startx - pa1.x) / dxa) * dya) + pa1.y;
			eya = (((endx - pa1.x) / dxa) * dya) + pa1.y;
		}
		
		if(Math.abs(dxb) < 0.01)
		{
			syb = pb1.y;
			eyb = pb2.y;
		}
		else
		{
			syb = (((startx - pb1.x) / dxb) * dyb) + pb1.y;
			eyb = (((endx - pb1.x) / dxb) * dyb) + pb1.y;
		}
		
		// finally, we check for intersection: if the y-coordinates of the
		//  segments at the intersection regions swap places (eg: segment 1 was
		//  higher at the beginning but lower than segment 2 in the end) then an
		//  intersection must take place
		
		if((sya > syb && eya < eyb) || (sya < syb && eya > eyb))
			return true;
		else
			return false; 
		
	}
	
	public static boolean intersectsContour(Point p1, Point p2, Contour src, int skip)
	{
		if(src == null)
			return false;
		
		Point prev;
		Point cur;
		
		int numPoints = src.path.size();
		
		// we need at least two points on the contour for it to make any sense
		if(numPoints < 2)
			return false;
		
		cur = src.path.get(0);
		
		for(int i = 1; i < numPoints; i += skip)
		{
			prev = cur;
			cur = src.path.get(i);
			
			if(segmentIntersection(p1, p2, prev, cur))
				return true;
		}
		
		// finally, we check the first-last point segment (because otherwise
		//  it would not be checked)
		prev = cur;
		cur = src.path.get(0);
		
		if(segmentIntersection(p1, p2, prev, cur))
			return true;
		
		return false;
	}
	
	public static void resolveIntersections(ArrayList<Cut> cuts, Cut newCut)
	{
		Cut curCut;
		
		for(int i = 0; i < cuts.size(); ++i)
		{
			curCut = cuts.get(i);
			
			if(segmentIntersection(curCut.p1, curCut.p2, newCut.p1, newCut.p2))
			{
				// in case of intersection, shortest cut takes precedence
				if(curCut.length <= newCut.length)
				{
					return;
				}
				else
				{
					cuts.set(i, newCut);
				}
			}	
		}
		
		// if we get this far, then no intersection must have occured
		cuts.add(newCut);
	}
	
	public static ArrayList<BodyComponent> makeComponents(Contour src, ArrayList<Cut> cuts)
	{
		final boolean DEBUG = true;
		
		if(cuts == null || cuts.size() == 0)
			return new ArrayList<BodyComponent>();
		
		ArrayList<BodyComponent> ret = new ArrayList<BodyComponent>();
		LinkedList<Segment> segments = new LinkedList<Segment>();
		
		int numComponents = 0;
		int startCut, endCut;
		
		Cut curCut;
		Cut primary;
		
		BodyComponent oldComp, newComp;
		Segment seg1, seg2, seg3;
		
		Segment curSeg;
		
		int oldOwner;
		
		boolean errorsRemaining = true;
		boolean segmentFound;
		
		while(errorsRemaining)
		{
			errorsRemaining = false;
			ret = new ArrayList<BodyComponent>();
			segments = new LinkedList<Segment>();
			numComponents = 0;
			
			// we need to make the first cut by hand because it is not possible to
			//  have a single segment that loops around the full contour
			
			curCut = cuts.get(0);
			startCut = curCut.p1.arcPos;
			endCut = curCut.p2.arcPos;
			
			// by convention, we want the end pos to be larger than the start pos
			if(startCut > endCut)
			{
				int temp = endCut;
				endCut = startCut;
				startCut = temp;
			}
			
			seg1 = new Segment(startCut, endCut, 0);
			seg2 = new Segment(endCut, startCut, 1);
			
			segments.add(seg1);
			segments.add(seg2);
			
			numComponents = 2;
			
			oldComp = new BodyComponent();
			newComp = new BodyComponent();
			
			oldComp.primaryCut = null;
			newComp.primaryCut = curCut;
			
			oldComp.links.add(newComp);
			newComp.links.add(oldComp);
			
			ret.add(oldComp);
			ret.add(newComp);
			
			for(int i = 1; i < cuts.size(); ++i)
			{
				curCut = cuts.get(i);
				startCut = curCut.p1.arcPos;
				endCut = curCut.p2.arcPos;
				
				if(startCut > endCut)
				{
					int temp = endCut;
					endCut = startCut;
					startCut = temp;
				}
				
				segmentFound = false;
				
				for(ListIterator<Segment> j = segments.listIterator(); j.hasNext();)
				{
					curSeg = j.next();
					
					if(curSeg.containsPos(startCut) && curSeg.containsPos(endCut))
					{
						// there is one special case we must consider, and that is
						//   when both the segment and the cut within the segment
						//   cross over zero, in which case startCut needs to be
						//   the larger absolute position
						
						
						if((curSeg.endPos < curSeg.startPos) && (startCut <= curSeg.endPos) && (endCut >= curSeg.startPos))
						{
							int temp = startCut;
							startCut = endCut;
							endCut = temp;
						}
						
						
						oldOwner = curSeg.owner;
						
						seg1 = new Segment(curSeg.startPos, startCut, oldOwner);
						seg2 = new Segment(startCut, endCut, numComponents);
						seg3 = new Segment(endCut, curSeg.endPos, oldOwner);
						
						++numComponents;
						
						oldComp = ret.get(oldOwner);
						newComp = new BodyComponent();
						
						newComp.primaryCut = curCut;
						
						oldComp.links.add(newComp);
						newComp.links.add(oldComp);
						
						ret.add(newComp);
						
						j.remove();
						j.add(seg1);
						j.add(seg2);
						j.add(seg3);
						
						segmentFound = true;
						
						break;
					}
				}
				
				if(!segmentFound)
				{
					// now we must find two sements in which the cut's endpoints
					// occur and which both belonging to the same owner
					
					// (actually, we just need to find the owner of those two
					//  segments, which simplifies the problem somewhat: we
					//  just make a table which keeps track of how many 'hits'
					//  each owner has had, and the first owner to reach two
					//  'hits' is the winner)
					
					int[] hits = new int[numComponents];
					
					int winner = -1;
					
					for(ListIterator<Segment> j = segments.listIterator(); j.hasNext();)
					{
						curSeg = j.next();
						
						if(curSeg.containsPos(startCut))
							hits[curSeg.owner] += 1;
						
						if(curSeg.containsPos(endCut))
							hits[curSeg.owner] += 1;
							
						if(hits[curSeg.owner] >= 2)
						{
							winner = curSeg.owner;
							break;
						}
					}
					
					if(winner == -1)
					{
						//System.out.println("Gigantic error! Cut not possible at all!");
						break;
					}
					
					// now we get the so-called primary cut of the component
					// and insert the current cut right after the primary
					
					// first we remove the current cut (we still have cuts(i) in 
					//  cutCut)
					cuts.remove(i);
					
					// now we find the primary cut and insert after it
					primary = ret.get(winner).primaryCut;
					
					if(primary != null)
					{
						for(int j = 0; j < cuts.size(); ++j)
						{
							if(cuts.get(j).equals(primary))
							{
								cuts.add(j+1, curCut);
								break;
							}
						}
					}
					else
					{
						cuts.add(0, curCut);
					}
					
					errorsRemaining = true;
					break;
					
				}
				
				/*
				if(DEBUG)
				{
					System.out.println("Cut not found!");
					System.out.println("From: " + startCut + ", To: " + endCut);
					System.out.println("Segments:");
					
					for(ListIterator<Segment> j = segments.listIterator(); j.hasNext();)
					{
						curSeg = j.next();
						System.out.print("[ " + curSeg.startPos + ", " + curSeg.endPos + " ]");
					}
				}*/
				
			}
		}
		
		// lastly, we go through the segments to construct all the contours 
		//  for the body segments
		
		int srcLength = src.path.size();
		int goLength;
		int cPos;
		BodyComponent curComponent;
		
		for(ListIterator<Segment> j = segments.listIterator(); j.hasNext();)
		{
			curSeg = j.next();
			
			curComponent = ret.get(curSeg.owner);
			
			if(curSeg.endPos < curSeg.startPos)
			{
				goLength = curSeg.endPos + srcLength - curSeg.startPos + 1;
			}
			else
			{
				goLength = curSeg.endPos - curSeg.startPos + 1;
			}
			
			cPos = curSeg.startPos;
			
			for(int i = 0; i < goLength; ++i)
			{
				curComponent.contour.path.add(new Point(src.path.get(cPos)));
				
				cPos = (cPos + 1) % srcLength;
			}
		}
		
		return ret;
	}
	
	public static ArrayList<Cut> getCuts(ArrayList<Point> crits, Contour src, double salienceThreshold, double maxLength, FloatImage srci)
	{
		final int SKIP_AMOUNT = 10;
		
		if(crits == null)
			return null;
		
		int numCrits = crits.size();
		
		int cLength = src.path.size();
		int cld2 = cLength / 2;
		
		double minDist1, minDist2;
		int minLoc1, minLoc2;
		
		int curPos1, curPos2;
		
		double x1, x2, y1, y2, dx, dy, arcPos;
		
		int midx, midy;
		
		double dist;
		
		Point tgt, startPoint;
		
		double salience;
		
		int oppositeLocation;
		
		double[][] srcMap = srci.getChannelData(0);
		
		ArrayList<Cut> cuts = new ArrayList<Cut>();
		Cut curCut;
		
		for(int i = 0; i < numCrits; ++i)
		{
			oppositeLocation = (crits.get(i).arcPos + cld2) % cLength;
			
			minLoc1 = -1;
			minDist1 = 10000.0;
			minLoc2 = -1;
			minDist2 = 10000.0;
			
			curPos1 = crits.get(i).arcPos;
			curPos2 = oppositeLocation;
			
			startPoint = crits.get(i);
			
			x1 = startPoint.x;
			y1 = startPoint.y;
			
			for(int j = 0; j < cld2; ++j)
			{
				tgt = src.path.get(curPos1);
				
				x2 = tgt.x;
				y2 = tgt.y;
				
				dx = x2 - x1;
				dy = y2 - y1;
				
				midx = (int)((x1 + x2) / 2.0);
				midy = (int)((y1 + y2) / 2.0);
				
				dist = Math.sqrt((dx * dx) + (dy * dy));
				
				if(dist > 0)
					salience = ((double)j) / dist;
				else
					salience = -1.0;
				
				if(dist < minDist1 && salience > salienceThreshold)
				{
					//if(!intersectsContour(startPoint, tgt, src, SKIP_AMOUNT))
					if(srcMap[midx][midy] > 0.9 && !intersectsContour(startPoint, tgt, src, SKIP_AMOUNT))
					{
						minLoc1 = curPos1;
						minDist1 = dist;
					}
				}
				
				tgt = src.path.get(curPos2);
				
				x2 = tgt.x;
				y2 = tgt.y;
				
				dx = x2 - x1;
				dy = y2 - y1;
				
				dist = Math.sqrt((dx * dx) + (dy * dy));
				
				if(dist > 0)
					salience = ((double)(cld2 - j)) / dist;
				else
					salience = -1.0;
				
				if(dist < minDist2 && salience > salienceThreshold)
				{
					//if(!intersectsContour(startPoint, tgt, src, SKIP_AMOUNT))
					if(srcMap[midx][midy] > 0.9 && !intersectsContour(startPoint, tgt, src, SKIP_AMOUNT))
					{
						minLoc2 = curPos2;
						minDist2 = dist;
					}
				} 
				
				curPos1 = (curPos1 + 1) % cLength;
				curPos2 = (curPos2 + 1) % cLength;	
			}
			
			if(minLoc1 >= 0)
			{
				//x2 = src.path.get(minLoc1).x;
				//y2 = src.path.get(minLoc1).y;
				
				tgt = src.path.get(minLoc1);
				tgt.arcPos = minLoc1;
				curCut = new Cut(startPoint, tgt);
				
				if(curCut.length <= maxLength)
					resolveIntersections(cuts, curCut);
				
				//drawLine(dest, (int)x1, (int)x2, (int)y1, (int)y2, 1.0);
			}
			if(minLoc2 >= 0)
			{
				//x2 = src.path.get(minLoc2).x;
				//y2 = src.path.get(minLoc2).y;
				
				tgt = src.path.get(minLoc2);
				tgt.arcPos = minLoc2;
				curCut = new Cut(startPoint, tgt);
				
				if(curCut.length <= maxLength)
					resolveIntersections(cuts, curCut);
				
				//drawLine(dest, (int)x1, (int)x2, (int)y1, (int)y2, 1.0);
			}
		}
		
		return cuts;
	}
	
	public static void drawCuts(ArrayList<Point> crits, Contour src, double salienceThreshold, double maxLength, FloatImage dest, FloatImage srci)
	{
		// get cuts
		ArrayList<Cut> cuts = getCuts(crits, src, salienceThreshold, maxLength, srci);
		
		
		// draw cuts
		Cut curCut = null;
		
		for(int n = 0; n < cuts.size(); ++n)
		{
			curCut = cuts.get(n);
			
			drawLine(dest, (int)curCut.p1.x, (int)curCut.p2.x, (int)curCut.p1.y, (int)curCut.p2.y, 0.0);
		}
		
		
		ArrayList<BodyComponent> comps = makeComponents(src, cuts);
	
		if(comps.size() != cuts.size() + 1 && (comps.size() != 0 || cuts.size() != 0))
		{
			System.out.println("Mismatch!: " + comps.size() + ", " + cuts.size());
		}
		
		BodyComponent temp;
		
		for(int i = 0; i < comps.size(); ++i)
		{
			temp = comps.get(i);
			temp.drawLinks(dest);
		}
		
		double[][] oblue = dest.getChannelData(2);
		
		oblue[(int)src.path.get(0).x][(int)src.path.get(0).y] = 1.0;
	}
}