import java.awt.geom.*;
import java.util.Random;

public class PlayingField {
	// posts are numbered clockwise starting from the upper right corner
	public static final int[] posts = {0, 1, 2, 3};
	public static final int UP_LEFT = 0;
	public static final int UP_RIGHT = 1;
	public static final int DOWN_RIGHT = 2;
	public static final int DOWN_LEFT = 3;
	
	private static Random rand = new Random();
	private static Rectangle2D.Double bound = new Rectangle2D.Double();
	
	/**
	 * set the boundary of the playing field -- DEPRICATED!
	 */
	public static void setBoundary(double x, double y, double width, double height) {
		bound.setRect(x, y, width, height);
	}
	
	/**
	 * set the boundary of the playing field (upper left corner at 0,0)
	 */
	public static void setBoundary(double width, double height) {
		bound.setRect(0, 0, width, height);
	}
	
	/**
	 * get the boundary of the playing field
	 */
	public static Rectangle2D.Double getBoundary() {
		return bound;
	}
	
	/**
	 * get the largerst square inside the playing field
	 */
	public static double getSquareSize() {
		return Math.min(bound.width, bound.height);
	}
	
	/**
	 * check if a given point is contained in the playing field
	 */
	public static boolean isInBounds(Point2D p) {
		return bound.contains(p);
	}
	
	/**
	 * find the coordinates of a given post
	 */
	private static Point2D.Double getPostCoords(int postNum) {
		double x = bound.x;
		double y = bound.y;
		
		if(postNum == UP_RIGHT || postNum == DOWN_RIGHT)
			x += bound.width;
		if(postNum == DOWN_LEFT || postNum == DOWN_RIGHT)
			y += bound.height;
		
		return new Point2D.Double(x,y);
	}
	
	/**
	 * returns the angle at which a particle would see a given post
	 * measured in radians from 0 to 2Pi
	 */
	public static double angleToPost(Particle p, int postNum){
		Point2D.Double post = getPostCoords(postNum);
		double theta = calcAngle(p,post);		
		return normalizeAngle(theta - p.theta);
		//return normalizeAngle(theta);
	}
	
	/**
	 * returns the absolute angle on the post that would be seen
	 * measured in radians from 0 to 2Pi
	 * NOTE: assumes theta=0 is in the +x direction
	 */
	public static double angleOnPost(Particle p, int postNum) {
		Point2D.Double post = getPostCoords(postNum);
		return calcAngle(post,p);
	}
	
	/**
	 * returns the distance from a particle to a post
	 */
	public static double distanceToPost(Particle p, int postNum) {
		Point2D.Double post = getPostCoords(postNum);
		return p.distance(post);
	}
	
	/**
	 * create a random point within the boundaries of the playing field
	 */
	public static Point2D.Double randomPoint() {
		double x = bound.getX() + bound.getWidth()*rand.nextDouble();
		double y = bound.getY() + bound.getHeight()*rand.nextDouble();
		return new Point2D.Double(x,y);		
	}
	
	/**
	 * create a random point centered in the middle of the playing field
	 */
	public static Point2D.Double randomCenteredPoint() {
		double x = bound.getX() + bound.getWidth()/2;
		double y = bound.getY() + bound.getHeight()/2;
		x += randGaussian()*bound.getWidth()/10;
		y += randGaussian()*bound.getHeight()/10;
		return new Point2D.Double(x,y);
	}
	
	
	/**
	 * return random double, uniformly distributed from 0 to 1
	 */
	public static double randUniform() {
		return rand.nextDouble();
	}
	
	/**
	 * return random double from the gaussian distribution mean=0, stdev=1
	 */
	public static double randGaussian() {
		return rand.nextGaussian();
	}
	
	/**
	 * calculate the angle defined by two points relative to the x axis
	 * measured in radians from 0 to 2Pi
	 */
    public static double calcAngle(Point2D.Double p1, Point2D.Double p2)
    {
    	double dx = p2.x-p1.x;
    	double dy = p2.y-p1.y;
        double angle;

        angle = Math.atan2(dy, dx);

        return normalizeAngle(angle);
    }
    
    /**
     * Normalize an angle to be between 0 and 2Pi
     */
    public static double normalizeAngle(double angle) {
    	while(angle < 0) {
    		angle += 2*Math.PI;
    	}
    	
    	while(angle > 2*Math.PI) {
    		angle -= 2*Math.PI;
    	}
    	
    	return angle;
    }
    
    /**
     * Normalize a distance by dividing it by the diagonal of the bounding rectangle
     */
    public static double normalizeDistance(double dist) {
    	double diagonal = Math.sqrt(Math.pow(bound.height,2) + Math.pow(bound.width,2));
    	return dist/diagonal;
    }
}