import java.util.ArrayList;
import java.util.Collections;

class TrackingPost
{
	int postNum;
	double angleCode;
	int[] grayCode;
	double cx;
	double cy;
	double height;
	double distance;
}

public class PostIdentifyUtils
{
	public static final double 	WIDTH_RATIO_MIN = 0.5;
	public static final double 	WIDTH_RATIO_MAX = 2.0;
	
	public static final double MIN_ASPECT = 0.20;
	public static final double MAX_ASPECT = 0.45;
	
	public static final double CENTER_BAND_POS_MIN = 0.45;
	public static final double CENTER_BAND_POS_MAX = 0.60;
	
	public static final double DISTANCE_HEIGHT_CONVERSION = 90.86;
	
	public static final int NUM_BANDS = 3;
	
	public static final double MIN_WIDTH = 4.0;
	
	public static final double 	MIDBAND_FRAC = 0.53;
	public static final double 	CODE_START = 0.612;
	public static final double 	CODE_END = 0.9;
	public static final int		CODE_BITS = 7;
	public static final double 	BIT_WIDTH = 0.043;
	
	public static final double IMAGE_HALF_WIDTH = 160.0;
	
	public static final double MAX_OFFSET_FRAC = 0.2;
	
	public static final int MIN_STRIPE_WIDTH = 3;
	
	public static int greyToInt(int[] greycode) {
		int len = greycode.length;
		int[] natBits = new int[len];
		int toReturn = 0;

		natBits[0] = greycode[0];
		for (int i = 1 ; i < len ; i++) {
			natBits[i] = natBits[i-1] ^ greycode[i];
		}
		for (int i = 0 ; i < len ; i++) {
			if (natBits[i] == 1)
				toReturn += Math.pow(2, len - i - 1);
		}
		
		return toReturn;
	}

	public static ArrayList<TrackingPost> findPosts(FloatImage idImage, FloatImage bwImage, FloatImage displayImage)
	{
		ArrayList<TrackingPost> posts = new ArrayList<TrackingPost>();
		
		double[][] intensity = bwImage.getChannelData(0);
		
		// First, identify all contours
		OUtil.EdgeField efield = new OUtil.EdgeField();
		efield.makeEdgeField(idImage);
		
		ArrayList<OUtil.Contour> contours = efield.getOutlines();
		
		// Now make bounding boxes out of the contours
		ArrayList<OUtil.BoundingBox> boxes = new ArrayList<OUtil.BoundingBox>();
		
		for(int i = 0; i < contours.size(); ++i)
		{
			boxes.add(new OUtil.BoundingBox(contours.get(i)));
		}
		
		// Quit if we didn't get two or more boxes
		if(boxes.size() < 2)
			return posts;
		
		// Sort the bounding boxes
		Collections.sort(boxes);
		
		// Now attempt to find matches between bounding boxes...
		double cx1, cx2;
		double moffset;
		double wratio;
		OUtil.BoundingBox b1, b2;
		
		for(int i = 0; i < boxes.size() - 1; ++i)
		{
			b1 = boxes.get(i);
			cx1 = (b1.left + b1.right) / 2.0;
			
			for(int j = i+1; j < boxes.size(); ++j)
			{
				b2 = boxes.get(j);	
				cx2 = (b2.left + b2.right) / 2.0;
				
				// Check to make sure that neigher is below the min width
				if(b1.width < MIN_WIDTH || b2.width < MIN_WIDTH)
					continue;
				
				// Check to see that their centers are not too far off
				moffset = MAX_OFFSET_FRAC * Math.max(b1.width, b2.width);
				if(Math.abs(cx1 - cx2) > moffset)
					continue;
				
				// Check to see that their widths are not too dissimilar
				wratio = b1.width / b2.width;
				if(wratio < WIDTH_RATIO_MIN || wratio > WIDTH_RATIO_MAX)
					continue;
					
				// We've passed these checks, so assign the match and break
				//  to the next box to try to match
				b1.match = b2;
				break;
			}
		}
		
		// Now hopefully we've found some matches, so just draw them onto the
		//  display image (for now)...
		OUtil.BoundingBox tempBox;
		OUtil.BoundingBox matchBox;
		for(int i = 0; i < boxes.size(); ++i)
		{
			tempBox = boxes.get(i);
			if(tempBox.match != null)
			{
				matchBox = tempBox.match;
				
				int x1,x2,y1,y2;
				
				// draw the bounding boxes themselves...
				x1 = (int)tempBox.left;
				x2 = (int)tempBox.right;
				y1 = (int)tempBox.top;
				y2 = (int)tempBox.bottom;
				OUtil.drawFilledBox(displayImage, x1, y1, x2, y2, 1.0, 0.0, 0.0);
				
				x1 = (int)matchBox.left;
				x2 = (int)matchBox.right;
				y1 = (int)matchBox.top;
				y2 = (int)matchBox.bottom;
				OUtil.drawFilledBox(displayImage, x1, y1, x2, y2, 0.0, 1.0, 0.0);
				
				// draw the connecting line...
				x1 = (int)((tempBox.left + tempBox.right) / 2.0);
				x2 = (int)((matchBox.left + matchBox.right) / 2.0);
				y1 = (int)tempBox.bottom;
				y2 = (int)matchBox.top;
				
				OUtil.drawLine(displayImage, x1, x2, y1, y2, 1.0);
			}
		}
		
		// now we unmark all the boxes
		for(int i = 0; i < boxes.size(); ++i)
			boxes.get(i).unmark();
			
		// now we follow unmarked chains and make them into posts
		OUtil.BoundingBox curBox, startBox, endBox, chainBox, centerBox;
		double postTop, postBottom, postHeight, postWidth, postMidpoint;
		int postIdStart, postIdEnd, postIdX, postIdCount;
		boolean previousBlack, curBlack;
		for(int i = 0; i < boxes.size(); ++i)
		{
			curBox = boxes.get(i);
			
			// don't bother with marked boxes...
			if(curBox.marked)
				continue;
				
			// now follow the box chain until the end...
			startBox = curBox;
			chainBox = startBox;
			int chainLength = 1;
			while(chainBox.match != null)
			{
				chainBox.mark();
				chainBox = chainBox.match;
				++chainLength;
			}
				
			endBox = chainBox;
			
			// if our chain doesn't have the right number of bands, it isn't
			//  a post
			if(chainLength != NUM_BANDS)
				continue;
			
			TrackingPost curPost = new TrackingPost();
				
			// now we get the absolute top and bottom
			postTop = startBox.top;
			curPost.cy = postTop;
			postBottom = endBox.bottom;
			
			curPost.cx = (startBox.left + startBox.right) / 2.0;
			
			// find the post height
			postHeight = postBottom - postTop;
			curPost.height = postHeight;
			curPost.distance = (DISTANCE_HEIGHT_CONVERSION / curPost.height) * 100.0;
			
			// additionally, we should check the center post's position
			centerBox = startBox.match;
			double centerBoxY = (centerBox.bottom - centerBox.top) / 2.0;
			double centerBoxPos = (centerBoxY - postTop) / postHeight;
			
			/*
			if(centerBoxPos < CENTER_BAND_POS_MIN ||
				centerBoxPos > CENTER_BAND_POS_MAX)
				continue;
			*/
			
			// find the width
			postWidth = endBox.right - startBox.left;
			
			// make sure the aspect ratio is within range
			double aspect = postWidth / postHeight;
			
			if(aspect < MIN_ASPECT || aspect > MAX_ASPECT)
				continue;
			
			// find the midpoint
			postMidpoint = (postHeight * MIDBAND_FRAC) + postTop;
			
			postIdStart = (int)(startBox.top);
			if(postIdStart < 0)
				postIdStart = 0;
				
			postIdEnd = (int)postMidpoint;
			if(postIdEnd >= displayImage.getHeight())
				postIdEnd = displayImage.getHeight() - 1;
				
			int postIdMid = (postIdStart + postIdEnd) / 2;
			
			int idHalfWidth = (int)((startBox.right - startBox.left) / 4.0);
			postIdX = (int)((startBox.left + startBox.right) / 2.0);
			
			double topSum = 0.0, bottomSum = 0.0;
			
			// draw the line dividing top and bottom for reference...
			OUtil.drawLine(displayImage, (int)curPost.cx - 20, (int)curPost.cx + 20, postIdMid, postIdMid, 1.0);
			
			// we sum the top and bottom...
			// first top
			for(int u = postIdX - idHalfWidth; u < postIdX + idHalfWidth; ++u)
			{
				for(int v = postIdStart; v < postIdMid; ++v)
				{
					topSum += intensity[u][v];
				}
			}
			
			// bottom
			for(int u = postIdX - idHalfWidth; u < postIdX + idHalfWidth; ++u)
			{
				for(int v = postIdMid; v < postIdEnd; ++v)
				{
					bottomSum += intensity[u][v];
				}
			}
			
			curPost.postNum = 0;
			
			// interpret the binary
			if(topSum > 0.0)
				curPost.postNum += 2;
			if(bottomSum > 0.0)
				curPost.postNum += 1;
			
			/*
			// scan
			previousBlack = false;
			postIdCount = 0;
			
			// now we start at the top, scan to the bottom and see how
			//  many times we cross from black to white
			int numOfConsecutive = 0;
			for(int j = postIdStart; j < postIdEnd; ++j)
			{
				if(j < 0 || postIdX < 0 || j >= idImage.getHeight() ||
					postIdX >= idImage.getWidth())
				{
					// fuck you!
					break;
				}
				
				curBlack = (intensity[postIdX][j] < 0.0);
				if(!curBlack && previousBlack && numOfConsecutive > MIN_STRIPE_WIDTH) 
				{
					numOfConsecutive = 0;
					++postIdCount;
				} else {
					numOfConsecutive++;
				}
				previousBlack = curBlack;
			}*/
			
			//curPost.postNum = postIdCount;
			
			// now read the gray code bits
			curPost.grayCode = new int[CODE_BITS];
			double curBitPos = postTop + (CODE_START * postHeight);
			double bitIncrement = postHeight * BIT_WIDTH;
			for(int j = 0; j < CODE_BITS; ++j)
			{
				if(postIdX < 0 || postIdX >= idImage.getWidth() || curBitPos < 0.0
					|| curBitPos >= idImage.getHeight())
				{
					// fuck you!
					break;
				}
				
				if(intensity[postIdX][(int)curBitPos] > 0.0)
					curPost.grayCode[j] = 1;
				else
					curPost.grayCode[j] = 0;
				curBitPos += bitIncrement;
			}
			
			posts.add(curPost);
		}
		
		// now draw the posts as well as their heights
		TrackingPost curPost;
		java.awt.Color drawColor = new java.awt.Color(0.0f, 1.0f, 0.0f);
		int drawX, drawY;
		for(int i = 0; i < posts.size(); ++i)
		{
			curPost = posts.get(i);
			
			// yay! draw the post
			drawX = (int)curPost.cx;
			drawY = (int)curPost.cy;
			String postString = "P " + curPost.postNum + " D " + (int)(curPost.distance * 100); 
			
			TinyWriter.drawString(postString, drawColor, displayImage, drawX, drawY, 2);
			
			if(curPost.grayCode == null)
				continue;
			
			String codeString = "";
			for(int k = 0; k < CODE_BITS; ++k)
			{
				if(curPost.grayCode[k] == 0)
					codeString += "0";
				else
					codeString += "1";
			}
			
			curPost.angleCode = (double)greyToInt(curPost.grayCode);
			
			TinyWriter.drawString(""+greyToInt(curPost.grayCode), drawColor, displayImage, drawX, drawY+20,3);
			
		}
		
		return posts;
	}

}