public class SweepCalibration implements Calibrator
{
	private IntPoint[][] transformTable;
	private double[][] baseline;
	private FloatImage internalBuffer;
	
	private int frameCounter;
	private int maxFrame;
	
	private int imgWidth;
	private int imgHeight;
	
	private static final double THRESHOLD = 0.5;
	
	private int mode;
	
	private static final int MODE_INITIALIZE = 1;
	private static final int MODE_HORIZONTAL = 2;
	private static final int MODE_VERTICAL = 3;
	private static final int MODE_IDENTIFY = 4;
	private static final int MODE_DONE = 5;
	
	public SweepCalibration()
	{
		transformTable = null;
		baseline = null;
		
		frameCounter = 0;
		maxFrame = 0;
		
		imgWidth = 0;
		imgHeight = 0;
		
		internalBuffer = null;
		
		mode = MODE_INITIALIZE;
	}
	
	public FloatImage getNextImage()
	{
		switch(mode)
		{
		case MODE_IDENTIFY:
			makeAllWhite();
			break;
		case MODE_HORIZONTAL:
			makeHorizontal(frameCounter);
			break;
		case MODE_VERTICAL:
			makeVertical(frameCounter);
			break;
		case MODE_DONE:
			return null;
		}
		
		return internalBuffer;
	}
	
	public void processResults(FloatImage result)
	{
		if(result != null)
		{
			switch(mode)
			{
			case MODE_INITIALIZE:
				imgWidth = result.getWidth();
				imgHeight = result.getHeight();
				internalBuffer = new FloatImage(3, imgWidth, imgHeight);
				calculateBaseline(result);
				mode = MODE_IDENTIFY;
				break;
			case MODE_IDENTIFY:
				identifyRegion(result);
				
				// this is equivalent to log_2
				maxFrame = imgWidth - 1;
				frameCounter = 0;
				
				mode = MODE_HORIZONTAL;
				break;
			case MODE_HORIZONTAL:
				processPattern(frameCounter, result);
				frameCounter++;
				if(frameCounter > maxFrame)
				{
					maxFrame = imgHeight - 1;
					frameCounter = 0;
					mode = MODE_VERTICAL;
				}
				break;
			case MODE_VERTICAL:
				processPattern(frameCounter, result);
				frameCounter++;
				if(frameCounter > maxFrame)
				{
					mode = MODE_DONE;
				}
				break;
			case MODE_DONE:
				break;
			}
		}
	}
	
	public void identifyRegion(FloatImage src)
	{	
		double[][] red = src.getChannelData(0);
		double[][] green = src.getChannelData(1);
		double[][] blue = src.getChannelData(2);
		
		int numIdentified = 0;
		
		transformTable = new IntPoint[imgWidth][imgHeight];
		
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{
				if((red[i][j] + green[i][j] + blue[i][j] - baseline[i][j]) > THRESHOLD)
				{
					transformTable[i][j] = new IntPoint();
					transformTable[i][j].x = 0;
					transformTable[i][j].y = 0;
					transformTable[i][j].weight = THRESHOLD;
					
					numIdentified++;
				}
			}
		}
		
		System.out.println("Number of pixels identified: " + numIdentified);
	}
	
	public void makeHorizontal(int sweepPosition)
	{	
		double[][] red = internalBuffer.getChannelData(0);
		double[][] green = internalBuffer.getChannelData(1);
		double[][] blue = internalBuffer.getChannelData(2);
		
		double fillValue = 0.0;
		
		for(int i = 0; i < imgWidth; ++i)
		{	
			if(i >= sweepPosition)
				fillValue = 1.0;
			else
				fillValue = -1.0;
		
			for(int j = 0; j < imgHeight; ++j)
			{
				red[i][j] = fillValue;
				green[i][j] = fillValue;
				blue[i][j] = fillValue;
			}
		}
	}
	
	public void makeVertical(int sweepPosition)
	{	
		double[][] red = internalBuffer.getChannelData(0);
		double[][] green = internalBuffer.getChannelData(1);
		double[][] blue = internalBuffer.getChannelData(2);
		
		double fillValue = 0.0;
		
		for(int j = 0; j < imgHeight; ++j)
		{	
			if(j >= sweepPosition)
				fillValue = 1.0;
			else
				fillValue = -1.0;
		
			for(int i = 0; i < imgWidth; ++i)
			{
				red[i][j] = fillValue;
				green[i][j] = fillValue;
				blue[i][j] = fillValue;
			}
		}
	}
	
	public void processPattern(int sweepPosition, FloatImage src)
	{	
		double[][] red = src.getChannelData(0);
		double[][] green = src.getChannelData(1);
		double[][] blue = src.getChannelData(2);
		
		double gain = 0.0;
		
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{
				if(transformTable[i][j] != null)
				{
					gain = red[i][j] + green[i][j] + blue[i][j] - baseline[i][j];
					
					if(gain > transformTable[i][j].weight)
					{
						if(mode == MODE_HORIZONTAL)
							transformTable[i][j].x = sweepPosition;
						else
							transformTable[i][j].y = sweepPosition;
							
						transformTable[i][j].weight = gain;
					}
				}
			}
		}
	}
	
	public void makeAllWhite()
	{
		double[][] red = internalBuffer.getChannelData(0);
		double[][] green = internalBuffer.getChannelData(1);
		double[][] blue = internalBuffer.getChannelData(2);
		
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{
				red[i][j] = 1.0;
				green[i][j] = 1.0;
				blue[i][j] = 1.0;
			}
		}
	}
	
	public void calculateBaseline(FloatImage src)
	{
		baseline = new double[imgWidth][imgHeight];
		
		double[][] red = src.getChannelData(0);
		double[][] green = src.getChannelData(1);
		double[][] blue = src.getChannelData(2);
		
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{
				baseline[i][j] = red[i][j] + green[i][j] + blue[i][j];
			}
		}
	}
	
	public void expandArea()
	{
		IntPoint[][] temp = new IntPoint[imgWidth][imgHeight];
		
		int expansion = 0;
		
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{
				if(false && transformTable[i][j] == null)
				{
					temp[i][j] = null;
				}
				else if(transformTable[i][j] == null || (transformTable[i][j].x == 0 && transformTable[i][j].y == 0))
				{
					temp[i][j] = new IntPoint();
					temp[i][j].x = 0;
					temp[i][j].y = 0;
					
					if(i > 0 && transformTable[i-1][j] != null && (transformTable[i-1][j].x != 0 || transformTable[i-1][j].y != 0))
					{
						temp[i][j].x = transformTable[i-1][j].x;
						temp[i][j].y = transformTable[i-1][j].y;
						expansion++;
					}
					else if(i < imgWidth-1 && transformTable[i+1][j] != null && (transformTable[i+1][j].x != 0 || transformTable[i+1][j].y != 0))
					{
						temp[i][j].x = transformTable[i+1][j].x;
						temp[i][j].y = transformTable[i+1][j].y;
						expansion++;
					}
					else if(j > 0 && transformTable[i][j-1] != null && (transformTable[i][j-1].x != 0 || transformTable[i][j-1].y != 0))
					{
						temp[i][j].x = transformTable[i][j-1].x;
						temp[i][j].y = transformTable[i][j-1].y;
						expansion++;
					}
					else if(j < imgHeight-1 && transformTable[i][j+1] != null && (transformTable[i][j+1].x != 0 || transformTable[i][j+1].y != 0))
					{
						temp[i][j].x = transformTable[i][j+1].x;
						temp[i][j].y = transformTable[i][j+1].y;
						expansion++;
					}
				}
				else
				{
					temp[i][j] = new IntPoint();
					temp[i][j].x = transformTable[i][j].x;
					temp[i][j].y = transformTable[i][j].y;
				}
			}
		}
		
		System.out.println("Expanded " + expansion + " times");
		
		transformTable = temp;
	}
	
	public void buildReverseTable()
	{
		IntPoint[][] temp = new IntPoint[imgWidth][imgHeight];
		int ti, tj;
		
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{	
				if(transformTable[i][j] != null)
				{
					ti = transformTable[i][j].x;
					tj = transformTable[i][j].y;
					
					if(ti >= 0 && ti < imgWidth && tj >= 0 && tj < imgHeight)
					{
						temp[ti][tj] = new IntPoint();
						temp[ti][tj].x = i;
						temp[ti][tj].y = j;
						temp[ti][tj].weight = 0.0;
					}
				}
			}
		}
		
		transformTable = temp;
	}
	
	public IntPoint[][] getTransform()
	{
		buildReverseTable();
		return transformTable;
	}
}